@@ -2,22 +2,20 @@ from __future__ import annotations
 from multiprocessing.context import SpawnContext
 import sys
 import traceback
-import queue
-import pickle
 from multiprocessing import get_context
-from threading import Thread, Condition, Lock
+from threading import Thread, Condition
 
 from collections import defaultdict
 from typing import (
-    Any,
     Dict,
-    Optional,
+    Mapping,
     Set,
     Tuple,
     Callable,
     List,
     Union,
     Sequence,
+    TypeVar,
 )
 
 from . import events
@@ -48,14 +46,20 @@ def _collect_rules(
     return list(ids), b2a
 
 
+_T_Rule = TypeVar("_T_Rule", bound=IRule)
+
+
 def make_mp_spawn(
-    id2rule: Sequence[IRule],
+    id2rule: Sequence[_T_Rule],
     ids: Sequence[int],
     dry_run: bool,
     keep_going: bool,
-    callback: Callable[[IEvent], None],
+    callback: Callable[[IEvent[_T_Rule]], None],
     njobs: int,
 ):
+    """
+    callback must be thread safe
+    """
     if len(ids) == 0:
         return MakeSummary(total=0, update=0, skip=0, fail=0, discard=0)
 
@@ -77,7 +81,7 @@ def make_mp_spawn(
 
     # Check inter-process transferability
     rules = [id2rule[i] for i in ids]
-    sendable = _test_interproc_portabability(rules, ctx)
+    sendable = _test_interproc_portabability([r.method for r in rules], ctx)
     _log_sendable_stats(sendable)
     sendable = {ids[j]: sendable[j] for j in range(len(ids))}
 
@@ -94,9 +98,6 @@ def make_mp_spawn(
     stop = False
 
     cv = Condition()  # for the above state vars
-    cb_lock = Lock()  # used when callback()
-
-    event_q = ctx.Queue()
 
     # Add rules with no dependencies to the job queue
     for i in ids:
@@ -148,49 +149,26 @@ def make_mp_spawn(
 
             cv.notify_all()
 
-    def callback_(e: IEvent):
-        with cb_lock:
-            callback(e)
-
-    def event_q_handler(workers: Sequence[Thread]):
-        # workers: list[Thread]
-        while True:
-            if all(not t.is_alive() for t in workers):
-                return
-
-            try:
-                callback_(event_q.get(True, 1))
-            except queue.Empty:
-                pass
-            except Exception:
-                traceback.print_exc()
-
     args = (
         ctx,
         get_job,
         set_result,
-        event_q,
         id2rule,
         updated_ids,
         main_ids,
         sendable,
         dry_run,
-        callback_,
+        callback,
     )
 
-    threads = [Thread(target=worker, args=(*args, i)) for i in range(njobs)]
+    threads = [Thread(target=worker, args=args) for _ in range(njobs)]
 
     for t in threads:
         t.start()
 
-    thread_event_q_handler = Thread(target=event_q_handler, args=(threads,))
-    thread_event_q_handler.start()
-
     try:
         for t in threads:
             t.join()
-
-        # thread_event_q_handler.join()
     except Exception:
         pass
     finally:
@@ -211,17 +189,14 @@ def worker(
     ctx: SpawnContext,
     get_job: Callable[[], Union[int, None]],
     set_result: Callable[[int, Union[Result, None]], None],
-    event_q: queue.Queue[IEvent],  # for process only
-    id2rule: List[IRule],
+    id2rule: List[_T_Rule],
     updated_ids: Set[int],
     main_ids: Set[int],
-    sendable: Sequence[int],
+    sendable: Mapping[int, bool],
     dry_run: bool,
-    callback: Callable[[IEvent], None],
-    name: Any,
+    callback: Callable[[IEvent[_T_Rule]], None],
 ):
-    name = f"worker{name}"
-    with ctx.Pool(1, _init_event_q, (event_q,)) as pool:
+    with ctx.Pool(1) as pool:
         while True:
             i = get_job()
 
@@ -234,12 +209,12 @@ def worker(
 
             try:
                 par_updated = any(dep in updated_ids for dep in rule.deps)
-                args = (rule, dry_run, par_updated, i in main_ids)
+                args = (rule, dry_run, par_updated, i in main_ids, callback)
 
                 if sendable[i]:
-                    res = pool.apply(process_worker, args)
+                    res = process_rule(*args, pool)
                 else:
-                    res = process_rule(*args, callback)
+                    res = process_rule(*args, None)
             except (Exception, KeyboardInterrupt) as e:
                 traceback.print_exc()
                 callback(events.FatalError(rule, e))
@@ -250,57 +225,26 @@ def worker(
                     return
 
 
-_event_q: Optional[queue.Queue[IEvent]] = None  # used by worker Processes
-
-
-def _init_event_q(q: queue.Queue[IEvent]):
-    global _event_q
-    assert _event_q is None
-    _event_q = q
-
-
-def process_worker(
-    rule: IRule, dry_run: bool, par_updated: bool, is_main: bool
-):
-    def cb(e: IEvent):
-        assert _event_q is not None
-        _event_q.put(e)
-
-    return process_rule(rule, dry_run, par_updated, is_main, cb)
-
-
 def _test_interproc_portabability(
-    objs: List[Any], ctx: SpawnContext
+    objs: Sequence[object], ctx: SpawnContext
 ) -> List[bool]:
     n = len(objs)
     picklable = [True] * n
 
-    codes: List[Optional[bytes]] = [None] * n
-
-    sys.stderr.write("Checking picklability\n")
-
-    for i in range(n):
-        try:
-            codes[i] = pickle.dumps(objs[i])
-        except Exception:
-            picklable[i] = False
-
-    sys.stderr.write("Checking inter-process portability\n")
+    sys.stderr.write("Checking inter-process transferability of rules\n")
 
     with ctx.Pool(1) as pool:
-        for i, code in enumerate(codes):
-            if code is not None:
-                picklable[i] = pool.apply(_test_transferability, (code,))
+        for i, obj in enumerate(objs):
+            try:
+                pool.apply(_dummy_func, (obj,))
+            except Exception:
+                picklable[i] = False
 
     return picklable
 
 
-def _test_transferability(pickle_code: bytes) -> bool:
-    try:
-        pickle.loads(pickle_code)
-        return True
-    except pickle.UnpicklingError:
-        return False
+def _dummy_func(_: object) -> bool:
+    ...
 
 
 def _log_sendable_stats(sendables: Sequence[bool]):
