@@ -7,7 +7,8 @@ def print_method(rule: IRule):
     info = rule._get_info()  # pyright: ignore [reportPrivateUsage]
     raw_rule = info.rule_store.rules[rule.raw_rule_id]
     sl = []
-    tostrs_func_call(sl, raw_rule.method, raw_rule.args, raw_rule.kwargs)
+    method = raw_rule.method
+    tostrs_func_call(sl, method, method.args, method.kwargs)
 
     a = create_default_logwriter("debug")
     a.debug(*sl)
